mock_app_update_id = "sdk.provider.mock.ktea-update"
mock_injection_bundle_id = "sdk.provider.mock.injection"
mock_initial_injection_bundle_id = "sdk.provider.mock.initial-injection"
deletion_spec_id = 'sdk.provider.deletion-spec'
